@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 2007, 2008 Free Software Foundation, Inc.
+ * Copyright 2007, 2008, 2009 Free Software Foundation, Inc.
  * Written by Paolo Bonzini.
  *
  * This file is part of GNU Smalltalk.
@@ -79,63 +79,76 @@ struct tool {
   const char *script;
   const char *options;
   const char *force_opt;
+  mst_Boolean allow_other_arguments;
 };
 
-struct tool tools[] = {
+const struct tool tools[] = {
   {
     "gst-convert", "scripts/Convert.st",
     "-h|--help --version -q|--quiet -v|-V|--verbose -C|--class: -r|--rule: \
         -c|--category: -f|--format: -o|--output: -I|--image-file: \
-        --kernel-directory:",
-    NULL
+        -F|--output-format: --kernel-directory:",
+    NULL, true
   },
   {
     "gst-load", "scripts/Load.st",
     "-h|--help --version -q|--quiet -v|-V|--verbose -n|--dry-run -f|--force \
 	--start:: -t|--test -I|--image-file: --kernel-directory: \
 	-i|--rebuild-image",
-    NULL
+    NULL, true
   },
   {
     "gst-reload", "scripts/Load.st",
     "-h|--help --version -q|--quiet -v|-V|--verbose -n|--dry-run -f|--force \
 	--start:: -t|--test -I|--image-file: --kernel-directory: \
 	-i|--rebuild-image",
-    "--force"
+    "--force\0", true
   },
   {
     "gst-package", "scripts/Package.st",
     "-h|--help --version --load --no-load --no-install --uninstall --dist \
         --prepare --test -t|--target-directory: --list-files: --list-packages \
-        --srcdir: --distdir|--destdir: --copy --all-files --vpath \
-        -n|--dry-run -I|--image-file: --kernel-directory:",
-    NULL
+        --srcdir: --distdir|--destdir: --copy --all-files --vpath -n|--dry-run \
+        -I|--image-file: --kernel-directory: --update|--download ",
+    NULL, true
   },
   {
     "gst-sunit", "scripts/Test.st",
     "-h|--help --version -q|--quiet -v|-V|--verbose -f|--file: -p|--package: \
 	-I|--image-file: --kernel-directory:",
-    NULL
+    NULL, true
   },
   {
-    "gst-blox", "scripts/Browser.st",
+    "gst-browser", "scripts/Load.st",
     "-I|--image-file: --kernel-directory:",
-    NULL
+    "--dry-run\0--start\0Browser\0", false
+  },
+  {
+    "gst-blox", "scripts/Load.st",
+    "-I|--image-file: --kernel-directory:",
+    "--dry-run\0--start\0BLOXBrowser\0", false
   },
   {
     "gst-doc", "scripts/GenDoc.st",
     "-h|--help --version -p|--package: -f|--file: -I|--image-file: \
         -n|--namespace: -o|--output: --kernel-directory: -F|--output-format:",
-    NULL
+    NULL, true
   },
   {
     "gst-remote", "scripts/Remote.st",
     "-h|--help --version --daemon --server -p|--port: -f|--file: -e|--eval: \
  	-l|--login: --package: --start: --stop: --pid --kill --snapshot:: \
 	-I|--image-file: --kernel-directory:",
-    NULL
+    NULL, true
   },
-  { NULL, NULL, NULL, NULL }
+  {
+    "gst-profile", "scripts/Profile.st",
+    "-f|--file: -e|--eval: -o|--output: -h|--help --version \
+	--no-separate-blocks",
+    NULL, true
+  },
+
+  { NULL, NULL, NULL, NULL, false }
 };
 
 /* An option parser compatible with the one in the Getopt class.
@@ -154,6 +167,7 @@ struct long_option {
 
 char short_opts[1 << (sizeof (char) * 8)];
 struct long_option *long_opts;
+const struct tool *tool;
 
 void
 option_error (const char *s, ...)
@@ -244,6 +258,12 @@ setup_options (const char *str)
 void
 parse_option (int short_opt, const char *long_opt, const char *arg)
 {
+  if (!short_opt && !long_opt && !tool->allow_other_arguments)
+    {
+      option_error ("invalid argument '%s'", arg);
+      return;
+    }
+
   if (short_opt == 'I'
       || (long_opt && !strcmp (long_opt, "image-file")))
     {
@@ -338,7 +358,7 @@ parse_short_options (const char *name, const char *arg)
         }
 
       else /* if (have_arg == OPT_MANDATORY) */
-        option_error ("expected argument for option -%s", name[-1]);
+        option_error ("expected argument for option -%c", short_opt);
     }
 
   return 1;
@@ -474,7 +494,8 @@ main(int argc, const char **argv)
     else if (!strcmp (tools[i].name, program_name))
       break;
 
-  setup_options (tools[i].options);
+  tool = &tools[i];
+  setup_options (tool->options);
   parse_options (&argv[1]);
 
 #ifdef HAVE_FORK
@@ -482,27 +503,36 @@ main(int argc, const char **argv)
     fork_daemon ();
 #endif
 
-  if (tools[i].force_opt)
+  if (tool->force_opt)
     {
-      smalltalk_argv = alloca (sizeof (const char *) * (argc + 1));
-      smalltalk_argc = argc;
-      smalltalk_argv[0] = tools[i].force_opt;
-      memcpy (&smalltalk_argv[1], &argv[1], argc * sizeof (char *));
+      const char *p;
+      int n;
+      for (p = tool->force_opt, n = 0; *p; p += strlen (p) + 1)
+	n++;
+
+      smalltalk_argc = argc + n - 1;
+      smalltalk_argv = alloca (sizeof (char *) * smalltalk_argc);
+      for (p = tool->force_opt, n = 0; *p; p += strlen (p) + 1)
+	smalltalk_argv[n++] = p;
+      memcpy (&smalltalk_argv[n], &argv[1], argc * sizeof (char *));
     }
   else
     {
-      smalltalk_argv = argv + 1;
       smalltalk_argc = argc - 1;
+      smalltalk_argv = argv + 1;
     }
 
 #ifdef CMD_LN_S
-  setenv ("LN_S", CMD_LN_S, 0);
+  if (!getenv ("LN_S"))
+    setenv ("LN_S", CMD_LN_S, 0);
 #endif
 #ifdef CMD_INSTALL
-  setenv ("INSTALL", CMD_INSTALL, 0);
+  if (!getenv ("INSTALL"))
+    setenv ("INSTALL", CMD_INSTALL, 0);
 #endif
 #ifdef CMD_ZIP
-  setenv ("XZIP", CMD_ZIP, 0);
+  if (!getenv ("XZIP"))
+    setenv ("XZIP", CMD_XZIP, 0);
 #endif
 
   gst_set_var (GST_VERBOSITY, 0);
@@ -512,9 +542,9 @@ main(int argc, const char **argv)
   if (result != 0)
     exit (result < 0 ? 1 : result);
     
-  if (!gst_process_file (tools[i].script, GST_DIR_KERNEL_SYSTEM))
+  if (!gst_process_file (tool->script, GST_DIR_KERNEL_SYSTEM))
     fprintf (stderr, "%s: Couldn't open kernel file `%s': %s\n",
-	     executable_name, tools[i].script, strerror (errno));
+	     executable_name, tool->script, strerror (errno));
 
   gst_invoke_hook (GST_ABOUT_TO_QUIT);
   exit (0);
