@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2005,2006,2007,2008
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2005,2006,2007,2008,2009
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -54,21 +54,21 @@
 
 #include "gstpriv.h"
 #include "re.h"
+#include "pointer-set.h"
 
 #include "ffi.h"
 #include <ltdl.h>
 
-typedef struct symbol_type_map
-{
-  OOP *symbol;
-  cdata_type type;
-}
-symbol_type_map;
+#ifdef HAVE_GETGRNAM
+#include <grp.h>
+#endif
+#ifdef HAVE_GETPWNAM
+#include <pwd.h>
+#endif
 
 typedef struct cparam
 {
   union {
-    int intVal;
     long longVal;
     PTR ptrVal;
     float floatVal;
@@ -91,11 +91,12 @@ typedef struct cfunc_info
 {
   avl_node_t avl;
   const char *funcName;
-  p_void_func funcAddr;
+  void (*funcAddr) ();
+} cfunc_info;
 
-  OOP cacheCFuncDesc;		/* used to test cache validity */
-  int cacheVariadic;		/* Is the function called with variadic parms? */
-  int needPostprocessing;	/* Is the function called with string/bytearray parms? */
+typedef struct cfunc_cif_cache
+{
+  unsigned cacheGeneration;	/* Is the function called with variadic parms? */
   ffi_cif cacheCif;		/* cached ffi_cif representation */
 
   int types_size;
@@ -103,17 +104,41 @@ typedef struct cfunc_info
   cparam *args;
   ffi_type **types;
 }
-cfunc_info;
+cfunc_cif_cache;
 
-typedef struct gst_stat
+typedef struct gst_ffi_closure
+{
+  // This field must come first, since the address of this field will
+  // be the same as the address of the overall structure.  This is due
+  // to disabling interior pointers in the GC.
+  ffi_closure closure;
+  void *address;
+  OOP callbackOOP;
+  ffi_cif cif;
+  ffi_type *return_type;
+  ffi_type *arg_types[1];
+}
+gst_ffi_closure;
+
+struct gst_stat_struct
 {
   unsigned short st_mode;	/* protection */
   long st_size;			/* total size, in bytes */
   long st_aTime;		/* time of last access */
   long st_mTime;		/* time of last modification */
   long st_cTime;		/* time of last change */
+};
+
+typedef struct gst_stat
+{
+  OBJ_HEADER;
+  OOP st_mode;		/* protection */
+  OOP st_size;		/* total size, in bytes */
+  OOP st_aTime;		/* time of last access */
+  OOP st_mTime;		/* time of last modification */
+  OOP st_cTime;		/* time of last change */
 }
-gst_stat;
+*gst_stat;
 
 
 
@@ -129,14 +154,10 @@ static void bad_type (OOP class_oop,
 static void push_smalltalk_obj (OOP oop,
 				cdata_type cType);
 
-/* Returns a descriptor for the latest C function which has been
-   registered using _gst_define_cfunc with the name FUNCNAME.  Returns
-   NULL if there is no such function.  */
-extern cfunc_info *lookup_function (const char *funcName);
-
 /* Converts the return type as stored in RESULT to an OOP, based
-   on the RETURNTYPEOOP that is stored in the descriptor.  */
-static OOP c_to_smalltalk (cparam *result, OOP returnTypeOOP);
+   on the RETURNTYPEOOP that is stored in the descriptor.  #void is
+   converted to RECEIVEROOP.  */
+static OOP c_to_smalltalk (cparam *result, OOP receiverOOP, OOP returnTypeOOP);
 
 /* Converts the return type CTYPE, stored in a descriptor to a
    libffi type.  */
@@ -147,10 +168,10 @@ static void init_dld (void);
 
 /* Wrapper around lt_dlopenext that invokes gst_initModule if it is found
    in the library.  */
-PTR dld_open (const char *filename);
+static PTR dld_open (const char *filename);
 
-/* Callout to tests callins.  */
-static void test_callin (OOP oop);
+/* Callout to tests callins and callbacks.  */
+static void test_callin (OOP oop, int(*callback)(const char *));
 
 /* Callout to test the CString class */
 static void test_cstring (char **string);
@@ -158,22 +179,23 @@ static void test_cstring (char **string);
 /* Callout to test #cObjectPtr parameters */
 static void test_cobject_ptr (const void **string);
 
-/* Converts a Symbol to a SmallInteger so that it is stored in the
-   CFunctionDescriptor.  For the return type isReturn is passed as
-   true and a CType is alwso considered valid.  */
-static OOP classify_type_symbol (OOP symbolOOP,
-				 mst_Boolean isReturn);
-
 /* Return the errno on output from the last callout.  */
 static int get_errno (void);
 
 /* Encapsulate binary incompatibilities between various C libraries.  */
+static int my_stat_old (const char *name,
+		        struct gst_stat_struct * out);
 static int my_stat (const char *name,
-		    gst_stat * out);
+		    OOP out);
+#ifdef HAVE_LSTAT
+static int my_lstat_old (const char *name,
+		         struct gst_stat_struct * out);
 static int my_lstat (const char *name,
-		     gst_stat * out);
+		     OOP out);
+#endif
 static int my_putenv (const char *str);
 static int my_chdir (const char *str);
+static int my_chown (const char *file, const char *owner, const char *group);
 static int my_symlink (const char* oldpath, const char* newpath);
 static char *my_mkdtemp (char* template);
 static int my_mkdir (const char* name, int mode);
@@ -187,8 +209,14 @@ static const char *get_argv (int n);
 /* The binary tree of function names vs. function addresses.  */
 static cfunc_info *c_func_root = NULL;
 
-/* The cfunc_info that's being called.  */
-static cfunc_info *c_func_cur = NULL;
+/* The binary tree of function names vs. function addresses.  */
+static struct pointer_map_t *cif_cache = NULL;
+
+/* Used to invalidate the cache upon GC.  */
+static unsigned cif_cache_generation = 1;
+
+/* The cfunc_cif_cache that's being filled in.  */
+static cfunc_cif_cache *c_func_cur = NULL;
 
 /* printable names for corresponding C types */
 static const char *c_type_name[] = {
@@ -225,41 +253,6 @@ static const char *c_type_name[] = {
   "char *",			/* CDATA_SYMBOL_OUT */
 };
 
-/* A map between symbols and the cdata_type enum.  */
-static const symbol_type_map type_map[] = {
-  {&_gst_unknown_symbol, CDATA_UNKNOWN},
-  {&_gst_char_symbol, CDATA_CHAR},
-  {&_gst_uchar_symbol, CDATA_UCHAR},
-  {&_gst_short_symbol, CDATA_SHORT},
-  {&_gst_ushort_symbol, CDATA_USHORT},
-  {&_gst_string_symbol, CDATA_STRING},
-  {&_gst_string_out_symbol, CDATA_STRING_OUT},
-  {&_gst_symbol_symbol, CDATA_SYMBOL},
-  {&_gst_symbol_out_symbol, CDATA_SYMBOL_OUT},
-  {&_gst_byte_array_symbol, CDATA_BYTEARRAY},
-  {&_gst_byte_array_out_symbol, CDATA_BYTEARRAY_OUT},
-  {&_gst_boolean_symbol, CDATA_BOOLEAN},
-  {&_gst_int_symbol, CDATA_INT},
-  {&_gst_uint_symbol, CDATA_UINT},
-  {&_gst_long_symbol, CDATA_LONG},
-  {&_gst_ulong_symbol, CDATA_ULONG},
-  {&_gst_float_symbol, CDATA_FLOAT},
-  {&_gst_double_symbol, CDATA_DOUBLE},
-  {&_gst_long_double_symbol, CDATA_LONG_DOUBLE},
-  {&_gst_void_symbol, CDATA_VOID},
-  {&_gst_variadic_symbol, CDATA_VARIADIC},
-  {&_gst_variadic_smalltalk_symbol, CDATA_VARIADIC_OOP},
-  {&_gst_c_object_symbol, CDATA_COBJECT},
-  {&_gst_c_object_ptr_symbol, CDATA_COBJECT_PTR},
-  {&_gst_smalltalk_symbol, CDATA_OOP},
-  {&_gst_self_symbol, CDATA_SELF},
-  {&_gst_self_smalltalk_symbol, CDATA_SELF_OOP},
-  {&_gst_wchar_symbol, CDATA_WCHAR},
-  {&_gst_wstring_symbol, CDATA_WSTRING},
-  {&_gst_wstring_out_symbol, CDATA_WSTRING_OUT},
-  {NULL, CDATA_UNKNOWN}
-};
-
 /* The errno on output from a callout */
 int _gst_errno = 0;
 
@@ -281,15 +274,30 @@ get_errno (void)
   int old;
   old = _gst_errno;
   _gst_errno = 0;
-  return (old);
+
+  /* When we get one of these, we don't return an error.  However,
+     the primitive still fails and the file/socket is closed by the
+     Smalltalk code.  */
+  if (old == ESHUTDOWN || old == ECONNRESET
+      || old == ECONNABORTED || old == ENETRESET
+      || old == EPIPE)
+    return 0;
+  else
+    return (old);
 }
 
-int
-my_stat (const char *name,
-	 gst_stat * out)
+static inline int
+adjust_time (time_t t)
+{
+  return _gst_adjust_time_zone (t) - 86400 * 10957;
+}
+
+static inline int
+my_stat_old (const char *name,
+	     struct gst_stat_struct * out)
 {
   int result;
-  static struct stat statOut;
+  struct stat statOut;
 
   result = stat (name, &statOut);
   if (!result)
@@ -297,20 +305,41 @@ my_stat (const char *name,
       errno = 0;
       out->st_mode = statOut.st_mode;
       out->st_size = statOut.st_size;
-      out->st_aTime = _gst_adjust_time_zone (statOut.st_atime) - 86400 * 10957;
-      out->st_mTime = _gst_adjust_time_zone (statOut.st_mtime) - 86400 * 10957;
-      out->st_cTime = _gst_adjust_time_zone (statOut.st_ctime) - 86400 * 10957;
+      out->st_aTime = adjust_time (statOut.st_atime);
+      out->st_mTime = adjust_time (statOut.st_mtime);
+      out->st_cTime = adjust_time (statOut.st_ctime);
+    }
+  return (result);
+}
+
+int
+my_stat (const char *name,
+	 OOP out)
+{
+  int result;
+  struct stat statOut;
+
+  result = stat (name, &statOut);
+  if (!result)
+    {
+      gst_stat obj = (gst_stat) OOP_TO_OBJ (out);
+      errno = 0;
+      obj->st_mode = FROM_INT (statOut.st_mode);
+      obj->st_aTime = FROM_INT (adjust_time (statOut.st_atime));
+      obj->st_mTime = FROM_INT (adjust_time (statOut.st_mtime));
+      obj->st_cTime = FROM_INT (adjust_time (statOut.st_ctime));
+      obj->st_size = FROM_OFF_T (statOut.st_size);
     }
   return (result);
 }
 
 #ifdef HAVE_LSTAT
-int
-my_lstat (const char *name,
-	 gst_stat * out)
+static inline int
+my_lstat_old (const char *name,
+	      struct gst_stat_struct * out)
 {
   int result;
-  static struct stat statOut;
+  struct stat statOut;
 
   result = lstat (name, &statOut);
   if (!result)
@@ -318,14 +347,36 @@ my_lstat (const char *name,
       errno = 0;
       out->st_mode = statOut.st_mode;
       out->st_size = statOut.st_size;
-      out->st_aTime = _gst_adjust_time_zone (statOut.st_atime) - 86400 * 10957;
-      out->st_mTime = _gst_adjust_time_zone (statOut.st_mtime) - 86400 * 10957;
-      out->st_cTime = _gst_adjust_time_zone (statOut.st_ctime) - 86400 * 10957;
+      out->st_aTime = adjust_time (statOut.st_atime);
+      out->st_mTime = adjust_time (statOut.st_mtime);
+      out->st_cTime = adjust_time (statOut.st_ctime);
+    }
+  return (result);
+}
+
+int
+my_lstat (const char *name,
+	 OOP out)
+{
+  int result;
+  struct stat statOut;
+
+  result = lstat (name, &statOut);
+  if (!result)
+    {
+      gst_stat obj = (gst_stat) OOP_TO_OBJ (out);
+      errno = 0;
+      obj->st_mode = FROM_INT (statOut.st_mode);
+      obj->st_aTime = FROM_INT (adjust_time (statOut.st_atime));
+      obj->st_mTime = FROM_INT (adjust_time (statOut.st_mtime));
+      obj->st_cTime = FROM_INT (adjust_time (statOut.st_ctime));
+      obj->st_size = FROM_OFF_T (statOut.st_size);
     }
   return (result);
 }
 #else
 #define my_lstat my_stat
+#define my_lstat_old my_stat_old
 #endif
 
 int
@@ -381,18 +432,19 @@ my_opendir (const char *dir)
 }
 
 void
-test_callin (OOP oop)
+test_callin (OOP oop, int(*callback)(const char *))
 {
   OOP o, sel;
   double f;
+  int i;
   _gst_str_msg_send (oop, "printNl", NULL);
 
   o = _gst_string_to_oop ("abc");
   sel = _gst_symbol_to_oop ("printNl");
   _gst_str_msg_send (_gst_str_msg_send (o, ",", o, NULL), "printNl",
 		     NULL);
-  _gst_msg_sendf (NULL, "%s %s printNl", "this is a test");
-  _gst_msg_sendf (&f, "%f %i + %f", 3, 4.7);
+  i = callback ("this is a test");
+  _gst_msg_sendf (&f, "%f %i + %f", i, 4.7);
   printf ("result = %f\n", f);
 }
 
@@ -435,7 +487,10 @@ dld_open (const char *filename)
   lt_dlhandle handle;
   void (*initModule) (struct VMProxy *);
 
-  handle = lt_dlopenext (filename);
+  /* Not all shared libraries have .xyz extensions! */
+  handle = lt_dlopen (filename);
+  if (!handle)
+    handle = lt_dlopenext (filename);
   if (handle)
     {
       initModule = lt_dlsym (handle, "gst_initModule");
@@ -448,6 +503,52 @@ dld_open (const char *filename)
   return (NULL);
 #endif
 }
+
+struct search_path_stack {
+  char *saved_search_path;
+  struct search_path_stack *next;
+};
+
+struct search_path_stack *search_paths;
+
+mst_Boolean
+_gst_dlopen (const char *path, mst_Boolean module)
+{
+  PTR h = dld_open (path);
+  if (h && !module)
+    _gst_msg_sendf (NULL, "%v %o addLibraryHandle: %C",
+		    _gst_class_name_to_oop ("DLD"), h);
+  return !!h;
+}
+
+void
+_gst_dladdsearchdir (const char *dir)
+{
+  lt_dlinsertsearchdir (lt_dlgetsearchpath (), dir);
+}
+
+void
+_gst_dlpushsearchpath (void)
+{
+  struct search_path_stack *entry = xmalloc (sizeof (struct search_path_stack));
+  const char *path = lt_dlgetsearchpath ();
+  entry->saved_search_path = path ? strdup (path) : NULL;
+  entry->next = search_paths;
+  search_paths = entry;
+}
+
+void
+_gst_dlpopsearchpath (void)
+{
+  struct search_path_stack *path = search_paths;
+  if (!path)
+    return;
+
+  lt_dlsetsearchpath (path->saved_search_path);
+  search_paths = path->next;
+  free (path->saved_search_path);
+  free (path);
+}
 
 
 
@@ -456,7 +557,11 @@ init_dld (void)
 {
   char *modules;
   lt_dlinit ();
-  lt_dladdsearchdir (MODULE_PATH);
+
+  modules = _gst_relocate_path (MODULE_PATH);
+  lt_dladdsearchdir (modules);
+  free (modules);
+
   if ((modules = getenv ("SMALLTALK_MODULES")))
     lt_dladdsearchdir (modules);
 
@@ -473,6 +578,8 @@ _gst_init_cfuncs (void)
 {
   extern char *getenv (const char *);
 
+  cif_cache = pointer_map_create ();
+
   /* Access to command line args */
   _gst_define_cfunc ("getArgc", get_argc);
   _gst_define_cfunc ("getArgv", get_argv);
@@ -489,10 +596,13 @@ _gst_init_cfuncs (void)
 
   _gst_define_cfunc ("errno", get_errno);
   _gst_define_cfunc ("strerror", strerror);
-  _gst_define_cfunc ("stat", my_stat);
-  _gst_define_cfunc ("lstat", my_lstat);
+  _gst_define_cfunc ("stat", my_stat_old);
+  _gst_define_cfunc ("lstat", my_lstat_old);
+  _gst_define_cfunc ("stat_obj", my_stat);
+  _gst_define_cfunc ("lstat_obj", my_lstat);
   _gst_define_cfunc ("utime", _gst_set_file_access_times);
   _gst_define_cfunc ("chmod", chmod);
+  _gst_define_cfunc ("chown", my_chown);
 
   _gst_define_cfunc ("opendir", my_opendir);
   _gst_define_cfunc ("closedir", closedir);
@@ -563,8 +673,8 @@ _gst_define_cfunc (const char *funcName,
 
 
 
-cfunc_info *
-lookup_function (const char *funcName)
+PTR
+_gst_lookup_function (const char *funcName)
 {
   cfunc_info *cfi = c_func_root;
 
@@ -574,7 +684,7 @@ lookup_function (const char *funcName)
 
       cmp = strcmp(funcName, cfi->funcName);
       if (cmp == 0)
-	return cfi;
+	return (PTR) cfi->funcAddr;
       
       cfi = (cfunc_info *) (cmp < 0 ? cfi->avl.avl_left : cfi->avl.avl_right);
     }
@@ -583,17 +693,88 @@ lookup_function (const char *funcName)
 }
 
 
+int
+_gst_c_type_size (int type)
+{
+  switch (type)
+    {
+    case CDATA_CHAR:
+      return sizeof (char);
+    case CDATA_UCHAR:
+      return sizeof (unsigned char);
+
+    case CDATA_SHORT:
+      return sizeof (short);
+    case CDATA_USHORT:
+      return sizeof (unsigned short);
+
+    case CDATA_INT:
+      return sizeof (int);
+    case CDATA_UINT:
+      return sizeof (unsigned int);
+
+    case CDATA_LONG:
+      return sizeof (long);
+    case CDATA_ULONG:
+      return sizeof (unsigned long);
+
+    case CDATA_FLOAT:
+      return sizeof (float);
+    case CDATA_DOUBLE:
+      return sizeof (double);
+    case CDATA_LONG_DOUBLE:
+      return sizeof (long double);
+
+    case CDATA_OOP:
+      return sizeof (OOP);
+
+    case CDATA_WCHAR:
+      return sizeof (wchar_t);
+
+    case CDATA_WSTRING:
+      return sizeof (wchar_t *);
+
+    case CDATA_STRING:
+    case CDATA_STRING_OUT:
+    case CDATA_SYMBOL:
+    case CDATA_BYTEARRAY:
+    case CDATA_BYTEARRAY_OUT:
+    case CDATA_SYMBOL_OUT:
+      return sizeof (char *);
+
+    case CDATA_COBJECT:
+      return sizeof (void *);
+
+    case CDATA_COBJECT_PTR:
+      return sizeof (void **);
+
+    default:
+      return 0;
+    }
+}
+
+void
+_gst_invalidate_croutine_cache (void)
+{
+  /* May want to delete and recreate completely upon global GC,
+     and do the cheap invalidation only for scavenging?  For now,
+     we do the simplest thing.  Incrementing by 2 makes sure that
+     the generation number is never 0.  */
+  cif_cache_generation += 2;
+}
+
 OOP
 _gst_invoke_croutine (OOP cFuncOOP,
 		      OOP receiver,
 		      OOP *args)
 {
-  gst_cfunc_descriptor desc;
+  gst_c_callable desc;
   cdata_type cType;
   cparam result, *local_arg_vec, *arg;
-  void **ffi_arg_vec;
-  OOP oop;
+  void *funcAddr, **p_slot, **ffi_arg_vec;
+  OOP *argTypes, oop;
   int i, si, fixedArgs, totalArgs;
+  mst_Boolean haveVariadic, needPostprocessing;
   inc_ptr incPtr;
 
   incPtr = INC_SAVE_POINTER ();
@@ -602,28 +783,53 @@ _gst_invoke_croutine (OOP cFuncOOP,
   INC_ADD_OOP (cFuncOOP);
   INC_ADD_OOP (receiver);
 
-  desc = (gst_cfunc_descriptor) OOP_TO_OBJ (cFuncOOP);
-  if (IS_NIL (desc->cFunction))
+  funcAddr = cobject_value (cFuncOOP);
+  if (!funcAddr)
     return (NULL);
 
-  c_func_cur = (cfunc_info *) COBJECT_VALUE (desc->cFunction);
-  if (!c_func_cur)
-    return (NULL);
+  p_slot = pointer_map_insert (cif_cache, cFuncOOP);
+  if (!*p_slot)
+    *p_slot = xcalloc (1, sizeof (cfunc_cif_cache));
 
-  fixedArgs = TO_INT (desc->numFixedArgs);
+  desc = (gst_c_callable) OOP_TO_OBJ (cFuncOOP);
+  argTypes = OOP_TO_OBJ (desc->argTypesOOP)->data;
+
+  c_func_cur = *p_slot;
+  fixedArgs = NUM_INDEXABLE_FIELDS (desc->argTypesOOP);
   totalArgs = 0;
+  haveVariadic = needPostprocessing = false;
   for (si = i = 0; i < fixedArgs; i++)
     {
-      cType = TO_INT (desc->argTypes[i]);
-      if (cType == CDATA_VARIADIC || cType == CDATA_VARIADIC_OOP)
+      cType = IS_OOP (argTypes[i]) ? CDATA_COBJECT : TO_INT (argTypes[i]);
+      switch (cType)
 	{
+	case CDATA_VOID:
+	  break;
+
+	case CDATA_VARIADIC:
+	case CDATA_VARIADIC_OOP:
 	  oop = args[si++];
 	  totalArgs += NUM_WORDS (OOP_TO_OBJ (oop));
+	  haveVariadic = true;
+	  break;
+
+        case CDATA_SELF:
+	case CDATA_SELF_OOP:
+	  totalArgs++;
+	  break;
+
+        case CDATA_COBJECT_PTR:
+        case CDATA_WSTRING_OUT:
+        case CDATA_STRING_OUT:
+        case CDATA_BYTEARRAY_OUT:
+	  needPostprocessing = true;
+	  /* fall through */
+
+	default:
+	  totalArgs++;
+	  si++;
+	  break;
 	}
-      else if (cType == CDATA_SELF || cType == CDATA_SELF_OOP)
-	totalArgs++;
-      else if (cType != CDATA_VOID)
-	totalArgs++, si++;
     }
 
   ffi_arg_vec = (void **) alloca (totalArgs * sizeof (void *));
@@ -639,7 +845,6 @@ _gst_invoke_croutine (OOP cFuncOOP,
     }
 
   c_func_cur->arg_idx = 0;
-  c_func_cur->cacheVariadic = false;
 
   for (i = 0; i < totalArgs; i++)
     ffi_arg_vec[i] = &local_arg_vec[i].u;
@@ -647,7 +852,7 @@ _gst_invoke_croutine (OOP cFuncOOP,
   /* Push the arguments */
   for (si = i = 0; i < fixedArgs; i++)
     {
-      cType = TO_INT (desc->argTypes[i]);
+      cType = IS_OOP (argTypes[i]) ? CDATA_COBJECT : TO_INT (argTypes[i]);
       if (cType == CDATA_SELF || cType == CDATA_SELF_OOP)
 	push_smalltalk_obj (receiver,
 			    cType == CDATA_SELF ? CDATA_UNKNOWN : CDATA_OOP);
@@ -658,29 +863,28 @@ _gst_invoke_croutine (OOP cFuncOOP,
 
   /* If the previous call was done through the same function descriptor,
      the ffi_cif is already ok.  */
-  if (c_func_cur->cacheCFuncDesc != cFuncOOP)
+  if (c_func_cur->cacheGeneration != cif_cache_generation)
     {
       ffi_prep_cif (&c_func_cur->cacheCif, FFI_DEFAULT_ABI, totalArgs,
-                    get_ffi_type (desc->returnType),
+                    get_ffi_type (desc->returnTypeOOP),
 		    c_func_cur->types);
 
       /* For variadic functions, we cannot cache the ffi_cif because
 	 the argument types change every time.  */
-      if (!c_func_cur->cacheVariadic)
-        c_func_cur->cacheCFuncDesc = cFuncOOP;
+      if (!haveVariadic)
+	c_func_cur->cacheGeneration = cif_cache_generation;
     }
 
   errno = 0;
-  ffi_call (&c_func_cur->cacheCif, FFI_FN (c_func_cur->funcAddr),
-	    &result.u, ffi_arg_vec);
+  ffi_call (&c_func_cur->cacheCif, FFI_FN (funcAddr), &result.u, ffi_arg_vec);
 
   _gst_set_errno (errno);
-  desc = (gst_cfunc_descriptor) OOP_TO_OBJ (cFuncOOP);
-  oop = c_to_smalltalk (&result, desc->returnType);
+  desc = (gst_c_callable) OOP_TO_OBJ (cFuncOOP);
+  oop = c_to_smalltalk (&result, receiver, desc->returnTypeOOP);
   INC_ADD_OOP (oop);
 
   /* Fixup all returned string variables */
-  if (c_func_cur->needPostprocessing)
+  if (needPostprocessing)
     for (i = 0, arg = local_arg_vec; i < totalArgs; i++, arg++)
       {
         if (!arg->oop)
@@ -689,7 +893,7 @@ _gst_invoke_croutine (OOP cFuncOOP,
         switch (arg->cType)
           {
           case CDATA_COBJECT_PTR:
-	    SET_COBJECT_VALUE (arg->oop, arg->u.cObjectPtrVal.ptrVal);
+	    set_cobject_value (arg->oop, arg->u.cObjectPtrVal.ptrVal);
 	    continue;
 
           case CDATA_WSTRING_OUT:
@@ -777,69 +981,35 @@ get_ffi_type (OOP returnTypeOOP)
    }
 }
 
-#define SET_TYPE(t)  c_func_cur->types[c_func_cur->arg_idx++] = (t);
-
-void
-push_smalltalk_obj (OOP oop,
-		    cdata_type cType)
+ffi_type *
+smalltalk_to_c (OOP oop,
+		cparam *cp,
+		cdata_type cType)
 {
-  OOP class;
-  int i;
-  cparam *cp = &c_func_cur->args[c_func_cur->arg_idx];
+  OOP class = OOP_INT_CLASS (oop);
 
-  class = OOP_INT_CLASS (oop);
+  if (cType == CDATA_UNKNOWN)
+    cType =
+      (oop == _gst_true_oop || oop == _gst_false_oop) ? CDATA_BOOLEAN :
+      oop == _gst_nil_oop ? CDATA_COBJECT :
+      class == _gst_char_class ? CDATA_CHAR :
+      class == _gst_unicode_character_class ? CDATA_WCHAR :
+      class == _gst_byte_array_class ? CDATA_BYTEARRAY : 
+      is_a_kind_of (class, _gst_integer_class) ? CDATA_LONG :
+      is_a_kind_of (class, _gst_string_class) ? CDATA_STRING :
+      is_a_kind_of (class, _gst_unicode_string_class) ? CDATA_WSTRING :
+      is_a_kind_of (class, _gst_c_object_class) ? CDATA_COBJECT :
+      is_a_kind_of (class, _gst_float_class) ? CDATA_DOUBLE :
+      CDATA_OOP;
 
-  switch (cType)
-    {
-    case CDATA_UNKNOWN:
-      cType =
-	(oop == _gst_true_oop || oop == _gst_false_oop) ? CDATA_BOOLEAN :
-	oop == _gst_nil_oop ? CDATA_COBJECT :
-	class == _gst_char_class ? CDATA_CHAR :
-	class == _gst_unicode_character_class ? CDATA_WCHAR :
-	class == _gst_byte_array_class ? CDATA_BYTEARRAY : 
-	is_a_kind_of (class, _gst_integer_class) ? CDATA_LONG :
-	is_a_kind_of (class, _gst_string_class) ? CDATA_STRING :
-	is_a_kind_of (class, _gst_unicode_string_class) ? CDATA_WSTRING :
-	is_a_kind_of (class, _gst_c_object_class) ? CDATA_COBJECT :
-	is_a_kind_of (class, _gst_float_class) ? CDATA_DOUBLE :
-	CDATA_OOP;
-      break;
-
-    case CDATA_VARIADIC:
-      if (class == _gst_array_class)
-	{
-	  c_func_cur->cacheVariadic = true;
-	  for (i = 1; i <= NUM_WORDS (OOP_TO_OBJ (oop)); i++)
-	    push_smalltalk_obj (ARRAY_AT (oop, i), CDATA_UNKNOWN);
-	}
-      else
-        bad_type (class, cType);
-
-      return;
-
-    case CDATA_VARIADIC_OOP:
-      if (class == _gst_array_class)
-	{
-	  c_func_cur->cacheVariadic = true;
-	  for (i = 1; i <= NUM_WORDS (OOP_TO_OBJ (oop)); i++)
-	    push_smalltalk_obj (ARRAY_AT (oop, i), CDATA_OOP);
-	}
-      else
-        bad_type (class, cType);
-
-      return;
-    }
-
-  cp->oop = NULL;
+  memset (cp, 0, sizeof (cparam));
   cp->cType = cType;
 
   if (cType == CDATA_OOP)
     {
       cp->u.ptrVal = (PTR) oop;
       INC_ADD_OOP (oop);	/* make sure it doesn't get gc'd */
-      SET_TYPE (&ffi_type_pointer);
-      return;
+      return &ffi_type_pointer;
     }
 
   else if (is_a_kind_of (class, _gst_integer_class))
@@ -850,47 +1020,42 @@ push_smalltalk_obj (OOP oop,
 	case CDATA_ULONG:
 	  cp->u.longVal = TO_C_LONG (oop);
 #if LONG_MAX == 2147483647
-          SET_TYPE (&ffi_type_sint32);
+          return &ffi_type_sint32;
 #else
-          SET_TYPE (&ffi_type_sint64);
+          return &ffi_type_sint64;
 #endif
-	  return;
 
         case CDATA_INT:
+	  cp->u.longVal = (int) TO_C_INT (oop);
+	  return &ffi_type_sint;
+
 	case CDATA_UINT:
-	  cp->u.intVal = TO_C_INT (oop);
-	  SET_TYPE (&ffi_type_sint);
-	  return;
+	  cp->u.longVal = (unsigned int) TO_C_INT (oop);
+	  return &ffi_type_sint;
 
 	case CDATA_CHAR:
-	  cp->u.intVal = (char) TO_C_INT (oop);
-	  SET_TYPE (&ffi_type_sint);
-	  return;
+	  cp->u.longVal = (char) TO_C_INT (oop);
+	  return &ffi_type_sint;
 
 	case CDATA_UCHAR:
-	  cp->u.intVal = (unsigned char) TO_C_INT (oop);
-	  SET_TYPE (&ffi_type_sint);
-	  return;
+	  cp->u.longVal = (unsigned char) TO_C_INT (oop);
+	  return &ffi_type_sint;
 
 	case CDATA_SHORT:
-	  cp->u.intVal = (short) TO_C_INT (oop);
-	  SET_TYPE (&ffi_type_sint);
-	  return;
+	  cp->u.longVal = (short) TO_C_INT (oop);
+	  return &ffi_type_sint;
 
 	case CDATA_USHORT:
-	  cp->u.intVal = (unsigned short) TO_C_INT (oop);
-	  SET_TYPE (&ffi_type_sint);
-	  return;
+	  cp->u.longVal = (unsigned short) TO_C_INT (oop);
+	  return &ffi_type_sint;
 
 	case CDATA_DOUBLE:
           cp->u.doubleVal = (double) TO_C_LONG (oop);
-	  SET_TYPE (&ffi_type_double);
-	  return;
+	  return &ffi_type_double;
 
 	case CDATA_FLOAT:
           cp->u.floatVal = (float) TO_C_LONG (oop);
-	  SET_TYPE (&ffi_type_float);
-	  return;
+	  return &ffi_type_float;
 	}
     }
 
@@ -902,11 +1067,10 @@ push_smalltalk_obj (OOP oop,
 	case CDATA_ULONG:
 	  cp->u.longVal = (oop == _gst_true_oop);
 #if LONG_MAX == 2147483647
-          SET_TYPE (&ffi_type_sint32);
+          return &ffi_type_sint32;
 #else
-          SET_TYPE (&ffi_type_sint64);
+          return &ffi_type_sint64;
 #endif
-	  return;
 
         case CDATA_INT:
 	case CDATA_UINT:
@@ -915,9 +1079,8 @@ push_smalltalk_obj (OOP oop,
 	case CDATA_SHORT:
 	case CDATA_USHORT:
 	case CDATA_BOOLEAN:
-	  cp->u.intVal = (oop == _gst_true_oop);
-	  SET_TYPE (&ffi_type_sint);
-	  return;
+	  cp->u.longVal = (oop == _gst_true_oop);
+	  return &ffi_type_sint;
 	}
     }
 
@@ -925,9 +1088,8 @@ push_smalltalk_obj (OOP oop,
 	    && (cType == CDATA_CHAR || cType == CDATA_UCHAR || cType == CDATA_WCHAR))
            || (class == _gst_unicode_character_class && cType == CDATA_WCHAR))
     {
-      cp->u.intVal = CHAR_OOP_VALUE (oop);
-      SET_TYPE (&ffi_type_sint);
-      return;
+      cp->u.longVal = CHAR_OOP_VALUE (oop);
+      return &ffi_type_sint;
     }
 
   else if (((class == _gst_string_class || class == _gst_byte_array_class)
@@ -937,29 +1099,22 @@ push_smalltalk_obj (OOP oop,
                && (cType == CDATA_SYMBOL || cType == CDATA_STRING)))
     {
       cp->oop = oop;
-      INC_ADD_OOP (oop);
 
       if (cp->cType == CDATA_BYTEARRAY || cp->cType == CDATA_BYTEARRAY_OUT)
 	cp->u.ptrVal = _gst_to_byte_array (oop);
       else
         cp->u.ptrVal = (gst_uchar *) _gst_to_cstring (oop);
 
-      c_func_cur->needPostprocessing = true;
-      SET_TYPE (&ffi_type_pointer);
-      return;
+      return &ffi_type_pointer;
     }
 
   else if (class == _gst_unicode_string_class
            && (cType == CDATA_WSTRING || cType == CDATA_WSTRING_OUT))
     {
       cp->oop = oop;
-      INC_ADD_OOP (oop);
-
       cp->u.ptrVal = (gst_uchar *) _gst_to_wide_cstring (oop);
 
-      c_func_cur->needPostprocessing = true;
-      SET_TYPE (&ffi_type_pointer);
-      return;
+      return &ffi_type_pointer;
     }
 
   else if (is_a_kind_of (class, _gst_float_class))
@@ -968,18 +1123,15 @@ push_smalltalk_obj (OOP oop,
 	{
 	case CDATA_LONG_DOUBLE:
 	  cp->u.longDoubleVal = _gst_oop_to_float (oop);
-	  SET_TYPE (&ffi_type_longdouble);
-	  return;
+	  return &ffi_type_longdouble;
 
 	case CDATA_DOUBLE:
 	  cp->u.doubleVal = _gst_oop_to_float (oop);
-	  SET_TYPE (&ffi_type_double);
-	  return;
+	  return &ffi_type_double;
 
 	case CDATA_FLOAT:
 	  cp->u.floatVal = (float) _gst_oop_to_float (oop);
-	  SET_TYPE (&ffi_type_float);
-	  return;
+	  return &ffi_type_float;
 	}
     }
 
@@ -988,22 +1140,17 @@ push_smalltalk_obj (OOP oop,
       switch (cType)
 	{
 	case CDATA_COBJECT_PTR:
-          c_func_cur->needPostprocessing = true;
 
 	  /* Set up an indirect pointer to protect against the OOP
 	     moving during the call-out.  */
 	  cp->u.cObjectPtrVal.pPtrVal = &cp->u.cObjectPtrVal.ptrVal;
-	  cp->u.cObjectPtrVal.ptrVal = COBJECT_VALUE (oop);
+	  cp->u.cObjectPtrVal.ptrVal = cobject_value (oop);
 	  cp->oop = oop;
-          INC_ADD_OOP (oop);
-
-	  SET_TYPE (&ffi_type_pointer);
-	  return;
+	  return &ffi_type_pointer;
 
 	case CDATA_COBJECT:
-	  cp->u.ptrVal = COBJECT_VALUE (oop);
-	  SET_TYPE (&ffi_type_pointer);
-	  return;
+	  cp->u.ptrVal = cobject_value (oop);
+	  return &ffi_type_pointer;
 	}
     }
 
@@ -1018,17 +1165,61 @@ push_smalltalk_obj (OOP oop,
 	case CDATA_STRING_OUT:
 	case CDATA_SYMBOL:
 	  cp->u.ptrVal = NULL;
-	  SET_TYPE (&ffi_type_pointer);
-	  return;
+	  return &ffi_type_pointer;
 	}
     }
 
+  /* #cObject can pass every object with non-pointer indexed instance
+     variables.  */
+  if (cType == CDATA_COBJECT)
+    {
+      switch (CLASS_INSTANCE_SPEC (class) & ISP_INDEXEDVARS)
+	{
+	case GST_ISP_FIXED:
+	case GST_ISP_POINTER:
+	  break;
+
+	default:
+	  /* Byte indexed variables, pass the pointer through.  */
+	  cp->u.ptrVal = OOP_TO_OBJ (oop)->data + CLASS_FIXED_FIELDS (class);
+	  return &ffi_type_pointer;
+	}
+    }
 
   bad_type (class, cType);
+  return NULL;
+}
+
+void
+push_smalltalk_obj (OOP oop,
+		    cdata_type cType)
+{
+  if (cType == CDATA_VARIADIC || cType == CDATA_VARIADIC_OOP)
+    {
+      int i;
+      if (OOP_INT_CLASS (oop) != _gst_array_class)
+	{
+          bad_type (OOP_INT_CLASS (oop), cType);
+          return;
+	}
+
+      cType = (cType == CDATA_VARIADIC) ? CDATA_UNKNOWN : CDATA_OOP;
+      for (i = 1; i <= NUM_WORDS (OOP_TO_OBJ (oop)); i++)
+	push_smalltalk_obj (ARRAY_AT (oop, i), cType);
+    }
+  else
+    {
+      cparam *cp = &c_func_cur->args[c_func_cur->arg_idx];
+      ffi_type *type = smalltalk_to_c (oop, cp, cType);
+      if (cp->oop && !IS_NIL (cp->oop))
+	INC_ADD_OOP (cp->oop);
+      if (type)
+	c_func_cur->types[c_func_cur->arg_idx++] = type;
+    }
 }
 
 OOP
-c_to_smalltalk (cparam *result, OOP returnTypeOOP)
+c_to_smalltalk (cparam *result, OOP receiverOOP, OOP returnTypeOOP)
 {
   cdata_type returnType;
   OOP resultOOP;
@@ -1041,36 +1232,36 @@ c_to_smalltalk (cparam *result, OOP returnTypeOOP)
   switch (returnType)
     {
     case CDATA_VOID:
-      resultOOP = _gst_nil_oop;
+      resultOOP = receiverOOP;
       break;
 
     case CDATA_CHAR:
     case CDATA_UCHAR:
-      resultOOP = CHAR_OOP_AT ((gst_uchar) result->u.intVal);
+      resultOOP = CHAR_OOP_AT ((gst_uchar) result->u.longVal);
       break;
 
     case CDATA_WCHAR:
-      resultOOP = char_new ((wchar_t) result->u.intVal);
+      resultOOP = char_new ((wchar_t) result->u.longVal);
       break;
 
     case CDATA_BOOLEAN:
-      resultOOP = result->u.intVal ? _gst_true_oop : _gst_false_oop;
+      resultOOP = result->u.longVal ? _gst_true_oop : _gst_false_oop;
       break;
 
     case CDATA_INT:
-      resultOOP = FROM_C_INT ((int) result->u.intVal);
+      resultOOP = FROM_C_INT ((int) result->u.longVal);
       break;
 
     case CDATA_UINT:
-      resultOOP = FROM_C_UINT ((unsigned int) result->u.intVal);
+      resultOOP = FROM_C_UINT ((unsigned int) result->u.longVal);
       break;
 
     case CDATA_SHORT:
-      resultOOP = FROM_INT ((short) result->u.intVal);
+      resultOOP = FROM_INT ((short) result->u.longVal);
       break;
 
     case CDATA_USHORT:
-      resultOOP = FROM_INT ((unsigned short) result->u.intVal);
+      resultOOP = FROM_INT ((unsigned short) result->u.longVal);
       break;
 
     case CDATA_LONG:
@@ -1104,8 +1295,8 @@ c_to_smalltalk (cparam *result, OOP returnTypeOOP)
 	{
 	  if (IS_INT (returnTypeOOP))
 	    returnTypeOOP = _gst_nil_oop;
-	  resultOOP = _gst_c_object_new (result->u.ptrVal, returnTypeOOP,
-					 _gst_c_object_class);
+	  resultOOP = COBJECT_NEW (result->u.ptrVal, returnTypeOOP,
+				   _gst_c_object_class);
 	}
       else if (returnType == CDATA_STRING || returnType == CDATA_STRING_OUT)
 	{
@@ -1154,89 +1345,88 @@ bad_type (OOP class_oop,
     _gst_errorf ("Attempt to pass an instance of %O as a %s", class_oop,
 	         c_type_name[cType]);
 
-  _gst_show_backtrace ();
+  _gst_show_backtrace (stderr);
+}
+
+
+/* This function does the unmarshaling of the libffi arguments to Smalltalk,
+   and calls the block that is stored in the CCallbackDescriptor.  */
+
+static void
+closure_msg_send (ffi_cif* cif, void* result, void** args, void* userdata)
+{
+  gst_ffi_closure *closure = userdata;
+  OOP callbackOOP = closure->callbackOOP;
+  gst_c_callable desc;
+  int numArgs, i;
+  OOP *argsOOP, *argTypes, resultOOP;
+  cdata_type cType;
+  cparam cp;
+
+  desc = (gst_c_callable) OOP_TO_OBJ (callbackOOP);
+  numArgs = NUM_INDEXABLE_FIELDS (desc->argTypesOOP);
+  argsOOP = alloca (sizeof (OOP) * numArgs);
+
+  for (i = 0; i < numArgs; i++)
+    {
+      memcpy (&cp.u, args[i], sizeof (ffi_arg));
+      desc = (gst_c_callable) OOP_TO_OBJ (callbackOOP);
+      argTypes = OOP_TO_OBJ (desc->argTypesOOP)->data;
+      argsOOP[i] = c_to_smalltalk (&cp, _gst_nil_oop, argTypes[i]);
+    }
+
+  desc = (gst_c_callable) OOP_TO_OBJ (callbackOOP);
+  resultOOP = _gst_nvmsg_send (desc->blockOOP, NULL, argsOOP, numArgs);
+
+  desc = (gst_c_callable) OOP_TO_OBJ (callbackOOP);
+  cType = IS_OOP (desc->returnTypeOOP) ? CDATA_COBJECT : TO_INT (desc->returnTypeOOP);
+  if (cType != CDATA_VOID)
+    {
+      smalltalk_to_c (resultOOP, &cp, cType);
+      memcpy (result, &cp.u, sizeof (ffi_arg));
+    }
+}
+
+void
+_gst_make_closure (OOP callbackOOP)
+{
+  gst_c_callable desc;
+  OOP *argTypes;
+  void *code;
+  gst_ffi_closure *closure;
+  int numArgs, i;
+
+  if (cobject_value (callbackOOP))
+    return;
+
+  desc = (gst_c_callable) OOP_TO_OBJ (callbackOOP);
+  numArgs = NUM_INDEXABLE_FIELDS (desc->argTypesOOP);
+  argTypes = OOP_TO_OBJ (desc->argTypesOOP)->data;
+  closure = (gst_ffi_closure *) ffi_closure_alloc (
+    sizeof (gst_ffi_closure) + sizeof(ffi_type *) * (numArgs - 1), &code);
+
+  closure->address = closure;
+  closure->callbackOOP = callbackOOP;
+  closure->return_type = get_ffi_type (desc->returnTypeOOP);
+  for (i = 0; i < numArgs; i++)
+    closure->arg_types[i] = get_ffi_type (argTypes[i]);
+
+  ffi_prep_cif (&closure->cif, FFI_DEFAULT_ABI,
+		numArgs, closure->return_type, closure->arg_types);
+
+  ffi_prep_closure_loc (&closure->closure, &closure->cif, closure_msg_send,
+			closure, code);
+  set_cobject_value (callbackOOP, code);
+}
+
+void
+_gst_free_closure (OOP callbackOOP)
+{
+  gst_ffi_closure *exec_closure = cobject_value (callbackOOP);
+  ffi_closure_free (exec_closure->address); 
+  set_cobject_value (callbackOOP, NULL);
 }
 
-
-
-OOP
-_gst_make_descriptor (OOP funcNameOOP,
-		      OOP returnTypeOOP,
-		      OOP argsOOP)
-{
-  char *funcName;
-  cfunc_info *cfi;
-  p_void_func funcAddr;
-  int numArgs, i;
-  gst_cfunc_descriptor desc;
-  inc_ptr incPtr;
-  OOP cFunction, cFunctionName, descOOP;
-
-  funcName = (char *) _gst_to_cstring (funcNameOOP);
-  cfi = lookup_function (funcName);
-  funcAddr = cfi ? cfi->funcAddr : NULL;
-
-  if (argsOOP == _gst_nil_oop)
-    numArgs = 0;
-  else
-    numArgs = NUM_OOPS (OOP_TO_OBJ (argsOOP));
-
-  /* since these are all either ints or new objects, I'm not moving the
-     OOPs */
-  incPtr = INC_SAVE_POINTER ();
-
-  cFunction = COBJECT_NEW (cfi);
-  INC_ADD_OOP (cFunction);
-
-  cFunctionName = _gst_string_new (funcName);
-  INC_ADD_OOP (cFunctionName);
-
-  desc = (gst_cfunc_descriptor)
-    new_instance_with (_gst_c_func_descriptor_class, numArgs,
-		       &descOOP);
-
-  desc->cFunction = cFunction;
-  desc->cFunctionName = cFunctionName;
-  desc->numFixedArgs = FROM_INT (numArgs);
-  desc->returnType = classify_type_symbol (returnTypeOOP, true);
-  for (i = 1; i <= numArgs; i++)
-    desc->argTypes[i - 1] =
-      classify_type_symbol (ARRAY_AT (argsOOP, i), false);
-
-  xfree (funcName);
-  INC_RESTORE_POINTER (incPtr);
-  return (descOOP);
-}
-
-OOP
-classify_type_symbol (OOP symbolOOP,
-		      mst_Boolean isReturn)
-{
-  const symbol_type_map *sp;
-  char *symbolName;
-
-  for (sp = type_map; sp->symbol != NULL; sp++)
-    {
-      if (*sp->symbol == symbolOOP)
-	return (FROM_INT (sp->type));
-    }
-
-  if (isReturn)
-    {
-      if (is_a_kind_of (OOP_CLASS (symbolOOP), _gst_c_type_class))
-	return (symbolOOP);	/* this is the type we want! */
-    }
-
-  symbolName = _gst_to_cstring (symbolOOP);	/* yeah, yeah...but
-						   they have the same
-						   representation! */
-  _gst_errorf ("Unknown data type symbol: %s", symbolName);
-
-  xfree (symbolName);
-
-  return (FROM_INT (CDATA_UNKNOWN));
-}
-
 void
 _gst_set_errno(int errnum)
 {
@@ -1252,6 +1442,104 @@ _gst_set_errno(int errnum)
 }
 
 
+int
+my_chown (const char *file, const char *user, const char *group)
+{
+#if defined HAVE_CHOWN && defined HAVE_GETGRNAM && defined HAVE_GETPWNAM
+  static char *save_user, *save_group;
+  static uid_t save_uid;
+  static gid_t save_gid;
+  static int recursive_depth;
+
+  uid_t uid, gid;
+  if (!file && !user && !group)
+    {
+      recursive_depth--;
+      if (recursive_depth == 0)
+	{
+#if defined HAVE_SETGROUPENT && defined HAVE_ENDGRENT
+	  endgrent ();
+#endif
+#if defined HAVE_SETPASSENT && defined HAVE_ENDPWENT
+	  endpwent ();
+#endif
+	}
+
+      free (save_user);
+      free (save_group);
+      save_user = save_group = NULL;
+      return 0;
+    }
+
+  if (!file)
+    {
+      recursive_depth++;
+      if (recursive_depth == 1)
+	{
+#if defined HAVE_SETGROUPENT && defined HAVE_ENDGRENT
+	  setgroupent (1);
+#endif
+#if defined HAVE_SETPASSENT && defined HAVE_ENDPWENT
+	  setpassent (1);
+#endif
+	}
+    }
+
+  if (!user)
+    uid = -1;
+  else if (save_user && !strcmp (save_user, user))
+    uid = save_uid;
+  else
+    {
+      struct passwd *pw;
+      pw = getpwnam (user);
+      if (!pw)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      uid = pw->pw_uid;
+      if (recursive_depth)
+	{
+	  if (save_user)
+	    free (save_user);
+          save_user = strdup (user);
+	  save_uid = uid;
+	}
+    }
+
+  if (!group)
+    gid = -1;
+  else if (save_group && !strcmp (save_group, group))
+    gid = save_gid;
+  else
+    {
+      struct group *gr;
+      gr = getgrnam (group);
+      if (!gr)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      gid = gr->gr_gid;
+      if (recursive_depth)
+	{
+	  if (save_group)
+	    free (save_group);
+          save_group = strdup (group);
+	  save_gid = gid;
+	}
+    }
+
+  if (!file)
+    return 0;
+  else
+    return chown (file, uid, gid);
+#endif
+}
+
 /* TODO: check if this can be changed to an extern declaration and/or
    an AC_CHECK_DECLS test.  */
 
