@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2005,2006,2007,2008
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2005,2006,2007,2008,2009
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -391,9 +391,6 @@ static void install_method (OOP methodOOP);
    _gst_prepare_execution_environment.  */
 static OOP termination_method;
 
-/* Used to number the lines starting from 1.  */
-static int line_offset;
-
 /* Used to abort really losing compiles, jumps back to the top level
    of the compiler */
 static jmp_buf bad_method;
@@ -454,7 +451,7 @@ _gst_install_initial_methods (void)
   _gst_set_compilation_category (_gst_string_new ("private"));
   _gst_alloc_bytecodes ();
   _gst_compile_byte (EXIT_INTERPRETER, 0);
-  _gst_compile_byte (RETURN_CONTEXT_STACK_TOP, 0);
+  _gst_compile_byte (JUMP_BACK, 4);
 
   /* The zeros are primitive, # of args, # of temps, stack depth */
   termination_method = _gst_make_new_method (0, 0, 0, 0, _gst_nil_oop,
@@ -468,12 +465,12 @@ _gst_install_initial_methods (void)
 
   install_method (termination_method);
 
-  methodsForString = "\
-methodsFor: aCategoryString [\
-    \"Calling this method prepares the parser to receive methods \
-      to be compiled and installed in the receiver's method dictionary. \
-      The methods are put in the category identified by the parameter.\" \
-    <primitive: VMpr_Behavior_methodsFor> \
+  methodsForString = "\n\
+methodsFor: aCategoryString [\n\
+    \"Calling this method prepares the parser to receive methods \n\
+      to be compiled and installed in the receiver's method dictionary. \n\
+      The methods are put in the category identified by the parameter.\" \n\
+    <primitive: VMpr_Behavior_methodsFor> \n\
 ]";
   _gst_set_compilation_class (_gst_behavior_class);
   _gst_set_compilation_category (_gst_string_new ("compiling methods"));
@@ -677,10 +674,11 @@ _gst_execute_statements (tree_node temps,
   incPtr = INC_SAVE_POINTER ();
   INC_ADD_OOP (methodOOP);
 
-  _gst_bytecode_counter = _gst_primitives_executed =
-    _gst_self_returns = _gst_inst_var_returns = _gst_literal_returns =
-    0;
-  _gst_sample_counter = 0;
+  if (!_gst_raw_profile)
+    _gst_bytecode_counter = _gst_primitives_executed =
+      _gst_self_returns = _gst_inst_var_returns = _gst_literal_returns =
+      _gst_sample_counter = 0;
+
   startTime = _gst_get_milli_time ();
 #ifdef HAVE_GETRUSAGE
   getrusage (RUSAGE_SELF, &startRusage);
@@ -716,7 +714,14 @@ _gst_execute_statements (tree_node temps,
       save_execution = _gst_execution_tracing;
       if (_gst_execution_tracing == 1)
         _gst_execution_tracing = 0;
-      _gst_str_msg_send (_gst_last_returned_value, "printNl", NULL);
+      if (_gst_responds_to (_gst_last_returned_value,
+			    _gst_intern_string ("printNl"))
+          || _gst_responds_to (_gst_last_returned_value,
+			       _gst_does_not_understand_symbol))
+        _gst_str_msg_send (_gst_last_returned_value, "printNl", NULL);
+      else
+	printf ("%O\n", _gst_last_returned_value);
+
       _gst_execution_tracing = save_execution;
     }
   if (quiet || _gst_verbosity < 3)
@@ -797,8 +802,6 @@ _gst_compile_method (tree_node method,
   this_method_category = _gst_this_category;
   _gst_unregister_oop (_gst_latest_compiled_method);
   _gst_latest_compiled_method = _gst_nil_oop;
-  _gst_line_number (-1, true);
-  line_offset = method->location.first_line - 1;
 
   incPtr = INC_SAVE_POINTER ();
 
@@ -806,6 +809,18 @@ _gst_compile_method (tree_node method,
   _gst_push_new_scope ();
   inside_block = 0;
   selector = compute_selector (method->v_method.selectorExpr);
+
+  /* When we are reading from stdin, it's better to write line numbers where
+     1 is the first line *in the current doit*, because for now the prompt
+     does not include the line number.  This might change in the future.
+
+     Also, do not emit line numbers if the method has no statements.  */
+  if ((method->location.file_offset != -1 && _gst_get_cur_stream_prompt ())
+      || !method->v_method.statements)
+    _gst_line_number (method->location.first_line, LN_RESET);
+  else
+    _gst_line_number (method->location.first_line, LN_RESET | LN_ABSOLUTE);
+
   INC_ADD_OOP (selector);
 
   if (_gst_declare_tracing)
@@ -964,7 +979,7 @@ compile_expression (tree_node expr)
 void
 compile_simple_expression (tree_node expr)
 {
-  _gst_line_number (expr->location.first_line - line_offset, false);
+  _gst_line_number (expr->location.first_line, 0);
   switch (expr->nodeType)
     {
     case TREE_VARIABLE_NODE:
@@ -1193,7 +1208,7 @@ compile_statements (tree_node statementList,
 
   if (isBlock)
     {
-      _gst_line_number (statementList->location.first_line - line_offset, true);
+      _gst_line_number (statementList->location.first_line, LN_FORCE);
       inside_block++;
     }
 
@@ -1218,7 +1233,7 @@ compile_statements (tree_node statementList,
 
   if (isBlock)
     {
-      _gst_line_number (-1, false);
+      _gst_line_number (-1, 0);
       inside_block--;
     }
 
@@ -1491,12 +1506,12 @@ compile_while_loop (OOP selector,
   for (oldJumpAroundLen = finalJumpLen = 0; ; oldJumpAroundLen = jumpAroundLen)
     {
       finalJumpOfs = whileCondLen + 2 + oldJumpAroundLen + jumpAroundOfs;
-      finalJumpLen = (finalJumpOfs + finalJumpLen > 65536 ? 6 :
-		      finalJumpOfs + finalJumpLen > 256 ? 4 : 2);
-      finalJumpLen = (finalJumpOfs + finalJumpLen > 65536 ? 6 :
-		      finalJumpOfs + finalJumpLen > 256 ? 4 : 2);
-      jumpAroundLen = (jumpAroundOfs + finalJumpLen > 65536 ? 6 :
-		       jumpAroundOfs + finalJumpLen > 256 ? 4 : 2);
+      finalJumpLen = (finalJumpOfs + finalJumpLen >= 65536 ? 6 :
+		      finalJumpOfs + finalJumpLen >= 256 ? 4 : 2);
+      finalJumpLen = (finalJumpOfs + finalJumpLen >= 65536 ? 6 :
+		      finalJumpOfs + finalJumpLen >= 256 ? 4 : 2);
+      jumpAroundLen = (jumpAroundOfs + finalJumpLen >= 65536 ? 6 :
+		       jumpAroundOfs + finalJumpLen >= 256 ? 4 : 2);
       if (jumpAroundLen == oldJumpAroundLen)
         break;
     }
@@ -1543,12 +1558,12 @@ compile_repeat (tree_node receiver)
   repeatedLoopLen = _gst_bytecode_length (receiverExprCodes);
 
   repeatedLoopLen += 2;
-  finalJumpLen = (repeatedLoopLen > 65536 ? 6 :
-		  repeatedLoopLen > 256 ? 4 : 2);
-  finalJumpLen = (repeatedLoopLen + finalJumpLen > 65536 ? 6 :
-		  repeatedLoopLen + finalJumpLen > 256 ? 4 : 2);
-  finalJumpLen = (repeatedLoopLen + finalJumpLen > 65536 ? 6 :
-		  repeatedLoopLen + finalJumpLen > 256 ? 4 : 2);
+  finalJumpLen = (repeatedLoopLen >= 65536 ? 6 :
+		  repeatedLoopLen >= 256 ? 4 : 2);
+  finalJumpLen = (repeatedLoopLen + finalJumpLen >= 65536 ? 6 :
+		  repeatedLoopLen + finalJumpLen >= 256 ? 4 : 2);
+  finalJumpLen = (repeatedLoopLen + finalJumpLen >= 65536 ? 6 :
+		  repeatedLoopLen + finalJumpLen >= 256 ? 4 : 2);
 
   _gst_compile_and_free_bytecodes (receiverExprCodes);
   _gst_compile_byte (POP_STACK_TOP, 0);	/* we don't care about expr's
@@ -1905,7 +1920,7 @@ compile_assignments (tree_node varList)
     {
       tree_node varName = varList->v_list.value;
 
-      _gst_line_number (varList->location.first_line - line_offset, false);
+      _gst_line_number (varList->location.first_line, 0);
       if (!_gst_find_variable (&variable, varName))
 	{
           if (varName->v_list.next)
@@ -2836,10 +2851,13 @@ method_new (method_header header,
       else
 	continue;
 
-      MAKE_OOP_UNTRUSTED (blockOOP, IS_OOP_UNTRUSTED (methodOOP));
       block = (gst_compiled_block) OOP_TO_OBJ (blockOOP);
-      block->method = methodOOP;
-      block->literals = literals;
+      if (IS_NIL (block->method))
+	{
+	  MAKE_OOP_UNTRUSTED (blockOOP, IS_OOP_UNTRUSTED (methodOOP));
+	  block->method = methodOOP;
+	  block->literals = literals;
+	}
     }
 
   if (bytecodes != NULL)
