@@ -6,7 +6,7 @@
 
 /***********************************************************************
  *
- * Copyright 2005, 2006, 2007, 2008 Free Software Foundation, Inc.
+ * Copyright 2005, 2006, 2007, 2008, 2009 Free Software Foundation, Inc.
  * Written by Paolo Bonzini.
  *
  * This file is part of GNU Smalltalk.
@@ -761,8 +761,10 @@ parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
 	          result = _gst_execute_statements (NULL, stmt, UNDECLARED_NONE,
 						    true);
 
-	          assert (result);
-		  DICTIONARY_AT_PUT (class_var_dict, name, result);
+	          if (result)
+		    DICTIONARY_AT_PUT (class_var_dict, name, result);
+		  else
+		    _gst_had_error = true;
 		}
 
 	      if (token (p, 0) != ']')
