@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2008
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2008,2009
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -69,9 +69,9 @@
 #include <math.h>
 #include <float.h>
 #include <errno.h>
+#include <signal.h>
 #include <sys/stat.h>
 #include <limits.h>
-#include <poll.h>
 #include <ctype.h>
 #include <wchar.h>
 #include <dirent.h>
@@ -198,6 +198,11 @@
 #define PREFETCH_ADDR(x, k) \
   DO_PREFETCH ((x), 0, (k));
 
+/* Synchronization primitives.  */
+#if !defined HAVE_SYNC_BUILTINS && defined __GNUC__
+#define __sync_synchronize()	__asm__ ("" : : : "memory")
+#endif
+
 /* Kill a warning when using GNU C.  Note that this allows using
    break or continue inside a macro, unlike do...while(0) */
 #ifdef __GNUC__
@@ -392,68 +397,6 @@ enum {
   (oop)->flags |= F_EPHEMERON;
 
 
-/*
-      3                   2                   1 
-    1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
-   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-   |    # fixed fields             |      unused       |I| kind  |1|
-   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-
-   I'm moving it to bits 13-30 (it used to stay at bits 5-30),
-   allocating space for more flags in case they're needed.
-   If you change ISP_NUMFIXEDFIELDS you should modify Behavior.st too.
-   Remember to shift by ISP_NUMFIXEDFIELDS-1 there, since Smalltalk does
-   not see ISP_INTMARK!!
-
-   Keep these in sync with _gst_sizes, in dict.c.
-
-   FIXME: these should be exported in a pool dictionary.  */
-
-enum {
-  /* This is a shift count.  */
-  ISP_NUMFIXEDFIELDS = 13,
-
-  /* Set if the indexed instance variables are pointers.  */
-  ISP_FIXED = 0,
-  ISP_SCHAR = 32,
-  ISP_UCHAR = 34,
-  ISP_CHARACTER = 36,
-  ISP_SHORT = 38,
-  ISP_USHORT = 40,
-  ISP_INT = 42,
-  ISP_UINT = 44,
-  ISP_FLOAT = 46,
-  ISP_INT64 = 48,
-  ISP_UINT64 = 50,
-  ISP_DOUBLE = 52,
-  ISP_UTF32 = 54,
-  ISP_LAST_SCALAR = 54,
-  ISP_POINTER = 62,
-
-#if SIZEOF_OOP == 8
-  ISP_LONG = ISP_INT64,
-  ISP_ULONG = ISP_UINT64,
-  ISP_LAST_UNALIGNED = ISP_FLOAT,
-#else
-  ISP_LONG = ISP_INT,
-  ISP_ULONG = ISP_UINT,
-  ISP_LAST_UNALIGNED = ISP_USHORT,
-#endif
-
-  /* Set if the instances of the class have indexed instance
-     variables.  */
-  ISP_ISINDEXABLE = 32,
-
-  /* These represent the shape of the indexed instance variables of
-     the instances of the class.  */
-  ISP_INDEXEDVARS = 62,
-  ISP_SHAPE = 30,
-
-  /* Set to 1 to mark a SmallInteger.  */
-  ISP_INTMARK = 1
-};
-
-
 /* the current execution stack pointer */
 #ifndef ENABLE_JIT_TRANSLATION
 # define sp		_gst_sp
@@ -572,6 +515,12 @@ extern OOP _gst_nil_oop
 #define BYTES_TO_SIZE(bytes) \
   ((bytes) / sizeof (PTR))
 
+#ifdef __GNUC__
+#define no_opt(x)	({ __typeof__ ((x)) _result; \
+			 asm ("" : "=r" (_result) : "0" ((x))); _result; })
+#else
+#define no_opt(x)	(x)
+#endif
 
 /* integer conversions and some information on SmallIntegers.  */
 
@@ -595,7 +544,7 @@ extern OOP _gst_nil_oop
    because they are likely to be optimized.  */
 
 #if SIZEOF_OOP == 4
-# if !defined(WORDS_BIGENDIAN) && defined (HAVE_INET_SOCKETS)
+# if !defined(WORDS_BIGENDIAN) && defined (HAVE_SOCKETS)
 #  define BYTE_INVERT(x) htonl((x))
 # elif defined _OS_OSBYTEORDERPPC_H
 #  define BYTE_INVERT(x) OSReadSwapInt32(&(x), 0)
@@ -636,7 +585,6 @@ extern OOP _gst_nil_oop
 #endif
 
 #include "ansidecl.h"
-#include "signalx.h"
 #include "mathl.h"
 #include "socketx.h"
 #include "strspell.h"
@@ -669,6 +617,8 @@ extern OOP _gst_nil_oop
 #include "mpz.h"
 #include "print.h"
 #include "security.h"
+#include "real.h"
+#include "sockets.h"
 
 /* Include this last, it has the bad habit of #defining printf
    and this fools gcc's __attribute__ (format) */
