@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008,2009
  * Free Software Foundation, Inc.
  * Written by Steve Byrne and Paolo Bonzini.
  *
@@ -158,16 +158,14 @@ static void grow_memory_no_compact (size_t size);
    memory.  */
 static void reset_survivor_space (struct surv_space *space);
 
+/* Return whether the incremental collector is running.  */
+static inline mst_Boolean incremental_gc_running (void);
+
 /* Restart the incremental collector.  Objects before FIRSTOOP
    are assumed to be alive (currently the base of the OOP table is
    always passed, but you never know).  */
 static void reset_incremental_gc (OOP firstOOP);
 
-/* The incremental collector has done its job.  Update statistics,
-   and if it was also sweeping old objects, make it consider all
-   objects as alive.  */
-static void finished_incremental_gc ();
-
 /* Gather statistics.  */
 static void update_stats (unsigned long *last, double *between, double *duration);
 
@@ -475,7 +473,7 @@ alloc_oop_table (size_t size)
   _gst_mem.num_free_oops = size;
   _gst_mem.first_allocated_oop = _gst_mem.ot;
   _gst_mem.last_allocated_oop = _gst_mem.last_swept_oop = _gst_mem.ot - 1;
-  _gst_mem.highest_swept_oop = _gst_mem.ot;
+  _gst_mem.next_oop_to_sweep = _gst_mem.ot - 1;
 }
 
 mst_Boolean
@@ -895,12 +893,22 @@ heap_data *
 oldspace_nomemory (heap_data *h, size_t sz)
 {
   if (!_gst_gc_running)
-    {
-      _gst_global_gc (sz);
-      return _gst_mem.old;
-    }
+    _gst_global_gc (sz);
   else
-    return NULL;
+    {
+      /* Already garbage collecting, emergency growth just to satisfy
+	 tenuring necessities.  */
+      int grow_amount_to_satisfy_rate = _gst_mem.old->heap_limit
+           * (100.0 + _gst_mem.space_grow_rate) / 100;
+      int grow_amount_to_satisfy_threshold = 
+	   (sz + _gst_mem.old->heap_total)
+	   * 100.0 /_gst_mem.grow_threshold_percent;
+
+      _gst_mem.old->heap_limit = MAX (grow_amount_to_satisfy_rate,
+				      grow_amount_to_satisfy_threshold);
+    }
+
+  return _gst_mem.old;
 }
 
 #ifndef NO_SIGSEGV_HANDLING
@@ -996,6 +1004,8 @@ _gst_compact (size_t new_heap_limit)
     }
   else
     {
+      /* Do not copy garbage.  */
+      _gst_finish_incremental_gc ();
       _gst_mem.numCompactions++;
       update_stats (&stats.timeOfLastCompaction, NULL, NULL);
     }
@@ -1053,7 +1063,6 @@ void
 _gst_global_compact ()
 {
   _gst_global_gc (0);
-  _gst_finish_incremental_gc ();
   _gst_compact (0);
 }
 
@@ -1112,12 +1121,26 @@ _gst_global_gc (int next_allocation)
     {
       old_limit = MAX (old_limit, _gst_mem.old->heap_total);
 
-      /* Check if it's time to compact the heap.  */
+      /* if memory is still low, go all the way on sweeping */
       if UNCOMMON ((next_allocation + _gst_mem.old->heap_total)
-	    * 100.0 / old_limit > _gst_mem.grow_threshold_percent)
+		    * 100.0 / old_limit > _gst_mem.grow_threshold_percent)
         {
-          s = "done, heap compacted";
-          _gst_compact (0);
+          int target_limit;
+          _gst_finish_incremental_gc ();
+
+	  /* Check if it's time to compact the heap. Compaction make the most 
+	     sense if there were lots of garbage. And the heap limit is shrunk 
+	     to avoid excessive garbage accumulation in the next round */
+	  target_limit = MAX(_gst_mem.eden.totalSize,
+			     ((next_allocation + _gst_mem.old->heap_total)
+			      * (100.0 + _gst_mem.space_grow_rate)
+			      / _gst_mem.grow_threshold_percent));
+	  if (target_limit < old_limit)
+            {
+              s = "done, heap compacted";
+              _gst_compact (0);
+              grow_memory_no_compact (target_limit);
+            }
         }
 
       /* Check if it's time to grow the heap.  */
@@ -1160,6 +1183,9 @@ _gst_scavenge (void)
 {
   int oldBytes, reclaimedBytes, tenuredBytes, reclaimedPercent;
 
+  /* Check if oldspace had to be grown in emergency.  */
+  size_t prev_heap_limit = _gst_mem.old->heap_limit;
+ 
   /* Force a GC as soon as possible if we're low on OOPs or memory.  */
   if UNCOMMON (_gst_mem.num_free_oops < LOW_WATER_OOP_THRESHOLD
      || _gst_mem.old->heap_total * 100.0 / _gst_mem.old->heap_limit >
@@ -1235,10 +1261,23 @@ _gst_scavenge (void)
 
   _gst_invalidate_croutine_cache ();
   mourn_objects ();
+
+  /* If tenuring had to grow oldspace, do a global garbage collection
+     now.  */
+  if (_gst_mem.old->heap_limit > prev_heap_limit)
+    {
+      _gst_global_gc (0);
+      _gst_incremental_gc_step ();
+      return;
+    }
 }
 
 
-
+mst_Boolean
+incremental_gc_running ()
+{
+  return (_gst_mem.next_oop_to_sweep > _gst_mem.last_swept_oop);
+}
 
 void
 _gst_finish_incremental_gc ()
@@ -1247,37 +1286,48 @@ _gst_finish_incremental_gc ()
 
 #if defined (GC_DEBUG_OUTPUT)
   printf ("Completing sweep (%p...%p), validity flags %x\n", _gst_mem.last_swept_oop,
-          _gst_mem.highest_swept_oop, _gst_mem.live_flags);
+          _gst_mem.next_oop_to_sweep, _gst_mem.live_flags);
 #endif
 
-  PREFETCH_START (_gst_mem.highest_swept_oop, PREF_BACKWARDS | PREF_READ | PREF_NTA);
-  for (oop = _gst_mem.highest_swept_oop, firstOOP = _gst_mem.last_swept_oop;
-       --oop > firstOOP; oop->flags &= ~F_REACHABLE)
+  PREFETCH_START (_gst_mem.next_oop_to_sweep, PREF_BACKWARDS | PREF_READ | PREF_NTA);
+  for (oop = _gst_mem.next_oop_to_sweep, firstOOP = _gst_mem.last_swept_oop;
+       oop > firstOOP; oop--)
     {
       PREFETCH_LOOP (oop, PREF_BACKWARDS | PREF_READ | PREF_NTA);
-      if (!IS_OOP_VALID_GC (oop))
+      if (IS_OOP_VALID_GC (oop))
+	{
+	  maybe_release_xlat (oop);
+	  oop->flags &= ~F_REACHABLE;
+	}
+      else
         {
           _gst_sweep_oop (oop);
 	  _gst_mem.num_free_oops++;
-          _gst_mem.highest_swept_oop = oop;
           if (oop == _gst_mem.last_allocated_oop)
             _gst_mem.last_allocated_oop--;
         }
     }
 
-  finished_incremental_gc ();
+  _gst_mem.next_oop_to_sweep = oop;
+  _gst_finished_incremental_gc ();
 }
 
 void
-finished_incremental_gc (void)
+_gst_finished_incremental_gc (void)
 {
+  if (_gst_mem.live_flags & F_OLD)
+    return;
+
   _gst_mem.live_flags &= ~F_REACHABLE;
   _gst_mem.live_flags |= F_OLD;
 
-  _gst_mem.reclaimedBytesPerGlobalGC =
-    _gst_mem.factor * stats.reclaimedOldSpaceBytesSinceLastGlobalGC +
-    (1 - _gst_mem.factor) * _gst_mem.reclaimedBytesPerScavenge;
-
+  if (stats.reclaimedOldSpaceBytesSinceLastGlobalGC) 
+    {
+      _gst_mem.reclaimedBytesPerGlobalGC =
+        _gst_mem.factor * stats.reclaimedOldSpaceBytesSinceLastGlobalGC +
+        (1 - _gst_mem.factor) * _gst_mem.reclaimedBytesPerGlobalGC;
+      stats.reclaimedOldSpaceBytesSinceLastGlobalGC = 0;
+    }
 #ifdef ENABLE_JIT_TRANSLATION
   /* Go and really free the blocks associated to garbage collected
      native code.  */
@@ -1285,33 +1335,41 @@ finished_incremental_gc (void)
 #endif
 }
 
-void
+mst_Boolean
 _gst_incremental_gc_step ()
 {
   OOP oop, firstOOP;
   int i;
 
-  for (i = 0, oop = _gst_mem.highest_swept_oop,
-       firstOOP = _gst_mem.last_swept_oop;
-       i <= INCREMENTAL_SWEEP_STEP && --oop > firstOOP;
-       oop->flags &= ~F_REACHABLE)
-    {
-      if (--oop > firstOOP)
-        {
-	  finished_incremental_gc ();
-	  break;
-	}
+  if (!incremental_gc_running ())
+    return true;
 
-      if (!IS_OOP_VALID_GC (oop))
+  i = 0;
+  firstOOP = _gst_mem.last_swept_oop;
+  for (oop = _gst_mem.next_oop_to_sweep; oop > firstOOP; oop--)
+    {
+      if (IS_OOP_VALID_GC (oop))
+	{
+	  maybe_release_xlat (oop);
+	  oop->flags &= ~F_REACHABLE;
+	}
+      else
         {
-          i++;
           _gst_sweep_oop (oop);
   	  _gst_mem.num_free_oops++;
-          _gst_mem.highest_swept_oop = oop;
           if (oop == _gst_mem.last_allocated_oop)
             _gst_mem.last_allocated_oop--;
+	  if (++i == INCREMENTAL_SWEEP_STEP)
+	    {
+	      _gst_mem.next_oop_to_sweep = oop - 1;
+	      return false;
+	    }
         }
     }
+
+  _gst_mem.next_oop_to_sweep = oop;
+  _gst_finished_incremental_gc ();
+  return true;
 }
 
 void
@@ -1346,7 +1404,7 @@ reset_incremental_gc (OOP firstOOP)
   _gst_mem.first_allocated_oop = oop;
 
 #ifdef NO_INCREMENTAL_GC
-  _gst_mem.highest_swept_oop = _gst_mem.last_allocated_oop + 1;
+  _gst_mem.next_oop_to_sweep = _gst_mem.last_allocated_oop;
   _gst_finish_incremental_gc ();
 #else
   /* Skip high OOPs that are unallocated.  */
@@ -1354,7 +1412,7 @@ reset_incremental_gc (OOP firstOOP)
     _gst_sweep_oop (oop);
 
   _gst_mem.last_allocated_oop = oop;
-  _gst_mem.highest_swept_oop = oop + 1;
+  _gst_mem.next_oop_to_sweep = oop;
 #endif
 
   _gst_mem.last_swept_oop = _gst_mem.first_allocated_oop - 1;
@@ -1369,9 +1427,9 @@ reset_incremental_gc (OOP firstOOP)
 
 #if defined (GC_DEBUG_OUTPUT)
   printf ("Found unallocated at OOP %p, last allocated OOP %p\n"
-          "Highest OOP swept top to bottom %p, lowest swept bottom to top %p\n",
+          "Next OOP swept top to bottom %p, lowest swept bottom to top %p\n",
 	  _gst_mem.first_allocated_oop, _gst_mem.last_allocated_oop,
-	  _gst_mem.highest_swept_oop, _gst_mem.last_swept_oop);
+	  _gst_mem.next_oop_to_sweep, _gst_mem.last_swept_oop);
 #endif
 }
 
@@ -1379,8 +1437,6 @@ reset_incremental_gc (OOP firstOOP)
 void
 _gst_sweep_oop (OOP oop)
 {
-  _gst_mem.last_swept_oop = oop;
-
   if (IS_OOP_FREE (oop))
     return;
 
@@ -1439,7 +1495,13 @@ mourn_objects (void)
       /* Copy the buffer into an Array */
       array = new_instance_with (_gst_array_class, size, &processor->gcArray);
       _gst_copy_buffer (array->data);
-      _gst_async_signal (processor->gcSemaphore);
+      if (!IS_NIL (processor->gcSemaphore))
+        _gst_async_signal (processor->gcSemaphore);
+      else
+	{
+	  _gst_errorf ("Running finalizers before initialization.");
+	  abort ();
+	}
     }
 }
 
