@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008,2009
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -218,9 +218,10 @@ struct memory_space
 
   /* These are the pointer to the first allocated OOP since the last
      completed incremental GC pass, to the first free OOP, to the
-     last low OOP considered by the incremental sweeper.  */
+     last low OOP considered by the incremental sweeper, to the
+     first high OOP not considered by the incremental sweeper.  */
   OOP first_allocated_oop, last_allocated_oop, last_swept_oop,
-      highest_swept_oop;
+      next_oop_to_sweep;
 
   /* The active survivor space */
   struct surv_space *active_half;
@@ -304,8 +305,14 @@ extern void _gst_global_gc (int next_allocation)
 extern void _gst_global_compact (void) 
   ATTRIBUTE_HIDDEN;
 
-/* Mark, sweep & compact the old objects.  */
-extern void _gst_incremental_gc_step (void) 
+/* Sweep a bunch of old objects, return whether there are more.  */
+extern mst_Boolean _gst_incremental_gc_step (void) 
+  ATTRIBUTE_HIDDEN;
+
+/* The incremental collector has done its job.  Update statistics,
+   and if it was also sweeping old objects, make it consider all
+   objects as alive.  */
+extern void _gst_finished_incremental_gc (void) 
   ATTRIBUTE_HIDDEN;
 
 /* Finish the incremental sweep phase of the GC.  */
