@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008,2009
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -65,9 +65,6 @@
 #define SEEK_END                2
 #endif
 
-#define TIMER_REAL       SIGALRM
-#define TIMER_PROCESS    SIGVTALRM
-
 #ifndef HAVE_SYS_MMAN_H
 #define PROT_READ       0x1             /* Page can be read.  */
 #define PROT_WRITE      0x2             /* Page can be written.  */
@@ -117,6 +114,9 @@ extern void _gst_signal_after (int deltaMilli,
 extern void _gst_init_sysdep (void)
   ATTRIBUTE_HIDDEN;
 
+extern void _gst_init_sysdep_win32 (void)
+  ATTRIBUTE_HIDDEN;
+
 /* Arrange so that FUNC is called when I/O becomes possible on
    registered file descriptors, and set up FD to trigger the
    signal.  */
@@ -130,8 +130,12 @@ extern void _gst_set_file_interrupt (int fd,
 extern void _gst_debug (void)
   ATTRIBUTE_HIDDEN;
 
+/* Sleeps for US milliseconds.  */
+extern void _gst_usleep (int us)
+  ATTRIBUTE_HIDDEN;
+
 /* Returns the local time since midnight in milliseconds */
-extern unsigned _gst_get_milli_time (void)
+extern uint64_t _gst_get_milli_time (void)
   ATTRIBUTE_HIDDEN;
 
 /* Returns the time in seconds since midnight Jan 1, 1970 (standard UNIX
@@ -286,4 +290,8 @@ extern void _gst_osmem_decommit (PTR base,
 				 size_t size)
   ATTRIBUTE_HIDDEN;
 
+/* Synchronously wait for FD to have input on it.  */
+extern void _gst_wait_for_input (int fd)
+  ATTRIBUTE_HIDDEN;
+
 #endif /* GST_SYSDEP_H */
