@@ -124,7 +124,8 @@ _gst_make_method (YYLTYPE *location,
 		  tree_node selectorExpr,
 		  tree_node temporaries,
 		  tree_node attributes,
-		  tree_node statements)
+		  tree_node statements,
+		  int isOldSyntax)
 {
   tree_node result;
 
@@ -134,6 +135,7 @@ _gst_make_method (YYLTYPE *location,
   result->v_method.temporaries = temporaries;
   result->v_method.attributes = attributes;
   result->v_method.statements = statements;
+  result->v_method.isOldSyntax = isOldSyntax;
   return (result);
 }
 
@@ -340,6 +342,19 @@ _gst_make_string_constant (YYLTYPE *location,
   return (result);
 }
 
+tree_node
+_gst_make_deferred_binding_constant (YYLTYPE *location,
+				     tree_node varNode)
+{
+  tree_node result;
+
+  result = make_tree_node (location, TREE_CONST_EXPR);
+  result->v_const.constType = CONST_DEFERRED_BINDING;
+  result->v_const.val.aVal = varNode;
+
+  return (result);
+}
+
 tree_node
 _gst_make_oop_constant (YYLTYPE *location,
 		        OOP oval)
@@ -694,6 +709,11 @@ print_method_node (tree_node node,
   indent (level);
   printf ("statements: ");
   _gst_print_tree (node->v_method.statements, level + 12);
+  indent (level);
+  if (node->v_method.isOldSyntax)
+    printf ("old syntax\n");
+  else
+    printf ("new syntax\n");
 }
 
 static void
@@ -747,6 +767,16 @@ print_const_node (tree_node node,
       _gst_print_tree (node->v_const.val.aVal, level + 7);
       break;
 
+    case CONST_DEFERRED_BINDING:
+      printf ("deferred variable binding: ");
+      _gst_print_tree (node->v_const.val.aVal, level + 27);
+      break;
+
+    case CONST_BINDING:
+      printf ("variable binding: ");
+      _gst_print_tree (node->v_const.val.aVal, level + 18);
+      break;
+
     default:
       _gst_errorf ("Unknown constant type %d", node->v_const.constType);
     }
